//----------------------------------------------------------------------------
//
//  This file is part of seq24.
//
//  seq24 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  seq24 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with seq24; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//-----------------------------------------------------------------------------
#ifndef SEQ24_CONTROLLERS
#define SEQ24_CONTROLLERS

const char * c_controller_names[] = {

    "0  Bank Select", 
    "1  Modulation Wheel ",
    "2  Breath controller ",
    "3",
    "4  Foot Pedal ",
    "5  Portamento Time ",
    "6  Data Entry ",
    "7  Volume ",
    "8  Balance ",
    "9",
    "10  Pan position",
    "11  Expression ",
    "12  Effect Control 1 ",
    "13  Effect Control 2 ",
    "14",
    "15",
    "16  General Purpose Slider 1",
    "17  General Purpose Slider 2",
    "18  General Purpose Slider 3",
    "19  General Purpose Slider 4",
    "20",
    "21",
    "22",
    "23",
    "24",
    "25",
    "26",
    "27",
    "28",
    "29",
    "30",
    "31",
    "32  Bank Select (fine)",
    "33  Modulation Wheel (fine)",
    "34  Breath controller (fine)",
    "35",
    "36  Foot Pedal (fine)",
    "37  Portamento Time (fine)",
    "38  Data Entry (fine)",
    "39  Volume (fine)",
    "40  Balance (fine)",
    "41",
    "42  Pan position (fine)",
    "43  Expression (fine)",
    "44  Effect Control 1 (fine)",
    "45  Effect Control 2 (fine)",
    "46",
    "47",
    "48",
    "49",
    "50",
    "51",
    "52",
    "53",
    "54",
    "55",
    "56",
    "57",
    "58",
    "59",
    "60",
    "61",
    "62",
    "63",
    "64  Hold Pedal (on/off)",
    "65  Portamento (on/off)",
    "66  Sustenuto Pedal (on/off)",
    "67  Soft Pedal (on/off)",
    "68  Legato Pedal (on/off)",
    "69  Hold 2 Pedal (on/off)",
    "70  Sound Variation",
    "71  Sound Timbre",
    "72  Sound Release Time",
    "73  Sound Attack Time",
    "74  Sound Brightness",
    "75  Sound Control 6",
    "76  Sound Control 7",
    "77  Sound Control 8",
    "78  Sound Control 9",
    "79  Sound Control 10",
    "80  General Purpose Button 1 (on/off)",
    "81  General Purpose Button 2 (on/off)",
    "82  General Purpose Button 3 (on/off)",
    "83  General Purpose Button 4 (on/off)",
    "84",
    "85",
    "86",
    "87",
    "88",
    "89",
    "90",
    "91  Effects Level",
    "92  Tremulo Level",
    "93  Chorus Level",
    "94  Celeste Level",
    "95  Phaser Level",
    "96  Data Button increment",
    "97  Data Button decrement",
    "98  Non-registered Parameter (fine)",
    "99  Non-registered Parameter (coarse)",
    "100  Registered Parameter (fine)",
    "101  Registered Parameter (coarse)",
    "102",
    "103",
    "104",
    "105",
    "106",
    "107",
    "108",
    "109",
    "110",
    "111",
    "112",
    "113",
    "114",
    "115",
    "116",
    "117",
    "118",
    "119",
    "120  All Sound Off",
    "121  All Controllers Off",
    "122  Local Keyboard (on/off)",
    "123  All Notes Off",
    "124  Omni Mode Off",
    "125  Omni Mode On",
    "126  Mono Operation",
    "127  Poly Operation"
    
};



















#endif
